use rand::Rng;
use rand_distr::Distribution;

use crate::params::Params;
use crate::ast::{
    Expr,
    Statement,
    Quantifier,
    BaseType,
    GlobalBlock,
    Block,
    Literal,
    Declaration,
    DeclarationBuilder,
    Variable,
    VariableBuilder,
    BinaryOperator,
};

pub struct AstBuilder {
    params: Params,
    ast: GlobalBlock,
    name_counter: u64,
    rng: rand::rngs::ThreadRng,
    rng_float: FloatGenerator,
}

impl AstBuilder {
    pub fn from(params: Params) -> Self {
        Self {
            ast: GlobalBlock::default(),
            name_counter: 0,
            rng: rand::thread_rng(),
            rng_float: FloatGenerator::new(&params),
            params,
        }
    }

    pub fn generate(&mut self) {

        let mut p: f64;

        loop {
            p = self.rng.gen();

            if p < self.params.global_flow.end_generation {
                break;
            } else {
                let decl = self.gen_decl();
                self.ast.push(decl);
            }
        }
    }

    fn gen_decl(&mut self) -> Box<dyn Statement> {
        let t = BaseType::Real;
        let v = self.gen_variable_quantified(t, Quantifier::Const);

        Box::new(
            DeclarationBuilder::default()
                .variable(v)
                .assn(self.gen_literal(t))
                .build()
                .unwrap()
        )
    }

    fn gen_variable(&mut self, t: BaseType) -> Variable {
        VariableBuilder::default()
            .type_(t)
            .name(self.gen_name())
            .quantifer(Quantifier::Var)
            .build()
            .unwrap()
    }

    fn gen_variable_quantified(&mut self, t: BaseType, q: Quantifier) -> Variable {
        VariableBuilder::default()
            .type_(t)
            .name(self.gen_name())
            .quantifer(q)
            .build()
            .unwrap()
    }


    fn gen_name(&mut self) -> String {
        let s = format!("_{}", self.name_counter);
        self.name_counter += 1;
        s
    }

    //fn gen_expr(&mut self, type_: BaseType) -> Box<dyn Expr> {
    //}

    fn gen_literal(&mut self, t: BaseType) -> Literal {
        match t {
            BaseType::Int => {
                let r = rand_distr::Beta::new(0.01, 0.01).unwrap();
                let i: i32 = (r.sample(&mut self.rng) * (i32::MAX as f64)) as i32;
                Literal::Int(i)
            }
            BaseType::Real => {
                let f = self.rng_float.sample(&mut self.rng);
                Literal::Real(f)
            }
            BaseType::Never => panic!("Attempted to generate literal of type Never"),
            BaseType::Unset => panic!("Attempted to generate literal of type Unset"),
        }
    }
}

impl ToString for AstBuilder {
    fn to_string(&self) -> String {
        let mut s = format!("// Generated by {} v{}\n",
            env!("CARGO_PKG_NAME"),
            env!("CARGO_PKG_VERSION"),
        );

        s.push_str(&self.ast.to_string());
        s
    }
}

/// Generates values from regions of interest for a float
///
/// These are concentrated around f32::MIN, 0, and f32::MAX, though any f32 value is possible
struct FloatGenerator {
    distro_pos: rand_distr::Normal<f32>,
    distro_zero: rand_distr::Normal<f32>,
    distro_neg: rand_distr::Normal<f32>,
}

impl FloatGenerator {
    fn new(p: &Params) -> Self {
        FloatGenerator {
            distro_pos: rand_distr::Normal::new(1.0, p.dev.float_gen_distro_pos_stddiv).unwrap(),
            distro_zero: rand_distr::Normal::new(0.0, 3.0).unwrap(),
            distro_neg: rand_distr::Normal::new(1.0, p.dev.float_gen_distro_neg_stddiv).unwrap(),
        }
    }
}

impl Distribution<f32> for FloatGenerator {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> f32 {
        let p: f64 = rng.gen();

        if p < 0.33 {
            f32::MAX - self.distro_pos.sample(rng).abs()
        } else if p < 0.66 {
            self.distro_zero.sample(rng)
        } else {
            -1.0 * (f32::MAX - self.distro_neg.sample(rng).abs())
        }
    }
}
